from typing import List, Optional, Dict, Any
from ..constants import LOG_INFERENCE_URL
from .log_stream_inference import LogStreamInference
from ..api_key import AthinaApiKey
from ..request_helper import RequestHelper
from ..util.token_count_helper import get_token_usage_openai_completion


class LogOpenAiCompletionStreamInference(LogStreamInference, AthinaApiKey):
    def __init__(self,
                 prompt_slug: str,
                 prompt: str,
                 language_model_id: str,
                 response_time: Optional[int] = None,
                 context: Optional[Dict] = None,
                 environment: Optional[str] = 'production',
                 customer_id: Optional[str] = None,
                 customer_user_id: Optional[str] = None,
                 session_id: Optional[str] = None,
                 user_query: Optional[str] = None,
                 external_reference_id: Optional[str] = None,
                 custom_attributes: Optional[Dict] = None):
        """
        constructor for log stream inference
        :param prompt_slug: str - The slug of the prompt used for the inference.
        :param prompt: str - The prompt used for the inference.
        :param language_model_id: str - The model used for the inference.
        :param response_time: Optional[int] - The response time in milliseconds. Defaults to None.
        :param context: Optional[Dict] - A dictionary containing additional context information. Defaults to None.
        :param environment: Optional[str] - The environment in which the inference occurred. Defaults to production.
        :param customer_id: Optional[str] - The customer's unique identifier. Defaults to None.
        :param customer_user_id: Optional[str] - The customer's user identifier. Defaults to None.
        :param session_id: Optional[str] - The session identifier. Defaults to None.
        :param user_query: Optional[str] - The user's query or input. Defaults to None.
        :param external_reference_id: Optional[str] - The external reference id. Defaults to None.
        :param custom_attributes: Optional[Dict] - A dictionary containing custom attributes. Defaults to None.
        """
        super().__init__(
            prompt_slug=prompt_slug,
            response_time=response_time,
            context=context,
            environment=environment,
            customer_id=customer_id,
            customer_user_id=customer_user_id,
            session_id=session_id,
            user_query=user_query,
            external_reference_id=external_reference_id,
            custom_attributes=custom_attributes)
        self.prompt = prompt
        self.language_model_id = language_model_id
        self.response = ''

    def _get_text_from_stream_chunk(self, stream_chunk):
        """
        gets the text from the stream chunk
        """
        try:
            text = ''
            choices = stream_chunk.get('choices', [])
            if choices and len(choices) > 0 and 'text' in choices[0]:
                text = choices[0].get('text', {})

            return text
        except Exception as e:
            raise e

    def collect_stream_inference(self, response):
        """
        collects the inference from the log stream
        """
        try:
            for stream_chunk in response:
                self.response += self._get_text_from_stream_chunk(
                    stream_chunk)
        except Exception as e:
            raise e

    def collect_stream_inference_by_chunk(self, stream_chunk):
        """
        collects the inference from the log stream of openai chat completion chunk by chunk
        """
        try:
            self.response += self._get_text_from_stream_chunk(
                stream_chunk)
        except Exception as e:
            raise e

    def log_stream_inference(self):
        """
        logs the stream inference to the athina api server
        """
        try:
            prompt_tokens = self._get_prompt_tokens(
                prompt=self.prompt, language_model_id=self.language_model_id)
            completion_tokens = self._get_completion_tokens(
                response=self.response, language_model_id=self.language_model_id)
            if prompt_tokens is not None and completion_tokens is not None:
                total_tokens = prompt_tokens + completion_tokens
            else:
                total_tokens = None
            payload = {
                'prompt_slug': self.prompt_slug,
                'prompt': self.prompt,
                'language_model_id': self.language_model_id,
                'response': self.response,
                'response_time': self.response_time,
                'context': self.context,
                'environment': self.environment,
                'customer_id': str(self.customer_id) if self.customer_id is not None else None,
                'customer_user_id': str(self.customer_user_id) if self.customer_user_id is not None else None,
                'session_id': str(self.session_id) if self.session_id is not None else None,
                'user_query': str(self.user_query) if self.user_query is not None else None,
                'external_reference_id': str(self.external_reference_id) if self.external_reference_id is not None else None,
                'prompt_tokens': prompt_tokens,
                'completion_tokens': completion_tokens,
                'total_tokens': total_tokens,
                'custom_attributes': self.custom_attributes,
            }
            # Remove None fields from the payload
            payload = {k: v for k, v in payload.items() if v is not None}
            RequestHelper.make_post_request(endpoint=LOG_INFERENCE_URL, payload=payload, headers={
                'athina-api-key': LogOpenAiCompletionStreamInference.get_api_key(),
            })
        except Exception as e:
            raise e

    def _get_prompt_tokens(self, prompt: str, language_model_id: str):
        """
        gets the prompt tokens given the prompt for the openai chat model completion
        """
        try:
            tokens = get_token_usage_openai_completion(
                text=prompt, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    def _get_completion_tokens(self, response: str, language_model_id: str):
        """
        gets the completion tokens given the prompt response from the openai chat model completion
        """
        try:
            tokens = get_token_usage_openai_completion(
                text=response, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None



from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Union, Sequence
from uuid import UUID

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import (
    AgentAction,
    AgentFinish,
    BaseMessage,
    ChatMessage,
    LLMResult,
)
from langchain.schema.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    HumanMessage,
    SystemMessage,
)
from langchain.schema.document import Document

from .inference_logger import InferenceLogger
from .api_key import AthinaApiKey
from .util.token_count_helper import get_prompt_tokens_openai_chat_completion, get_completion_tokens_openai_chat_completion, get_token_usage_openai_completion


class CallbackHandler(BaseCallbackHandler, AthinaApiKey):
    """
    Callback handler for the LangChain API.
    """

    def __init__(
        self,
        prompt_slug: str,
        user_query: Optional[str] = None,
        environment: Optional[str] = 'production',
        session_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        external_reference_id: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the CallbackHandler"""
        if kwargs:
            self.global_context = kwargs
        else:
            self.global_context = None

        self.prompt_slug = prompt_slug
        self.user_query = user_query
        self.environment = environment
        self.session_id = session_id
        self.customer_id = customer_id
        self.customer_user_id = customer_user_id
        self.external_reference_id = external_reference_id
        self.custom_attributes = custom_attributes
        self.runs: Dict[UUID, Dict[str, Any]] = {}

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            if run_id is None or run_id not in self.runs:
                return
            run_info = self.runs[run_id]
            retrieved_documents_data = ''
            for document in documents:
                page_content = document.page_content
                retrieved_documents_data += page_content + '\n'
            self.global_context['documents'] = retrieved_documents_data
            run_info['retrieved_documents'] = self.global_context
        except Exception as e:
            exception_message = (
                f"Error:\n"
                f"service name: athina-logger\n"
                f"file name: langchain_handler\n"
                f"method name: on_retriever_end\n"
                f"{str(e)}"
            )
            print(exception_message)

    def on_chat_model_start(
        self, serialized: Dict[str, Any], messages: List[List[BaseMessage]], run_id: UUID, parent_run_id: Optional[UUID] = None, **kwargs: Any,
    ) -> Any:
        """
        Log the chat model start
        """
        try:
            for message in messages:
                message_dicts = self._create_message_dicts(message)

            self.runs[run_id] = {
                'is_chat_model': True,
                'prompt_slug': self.prompt_slug,
                'user_query': self.user_query,
                'context': self.global_context,
                'prompt': message_dicts,
                'session_id': self.session_id,
                'customer_id': self.customer_id,
                'customer_user_id': self.customer_user_id,
                'external_reference_id': self.external_reference_id,
                'custom_attributes': self.custom_attributes,
                'llm_start_time': datetime.now(timezone.utc),
                'language_model_id': kwargs.get('invocation_params').get('model_name')
            }
        except Exception as e:
            exception_message = (
                f"Error:\n"
                f"service name: athina-logger\n"
                f"file name: langchain_handler\n"
                f"method name: on_chat_model_start\n"
                f"{str(e)}"
            )
            print(exception_message)

    def on_llm_new_token(
        self,
        token: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""
        pass

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            self.runs[run_id] = {
                'is_chat_model': False,
                'prompt_slug': self.prompt_slug,
                'user_query': self.user_query,
                'context': self.global_context,
                'prompt': {'text': ' '.join(prompts)},
                'session_id': self.session_id,
                'customer_id': self.customer_id,
                'customer_user_id': self.customer_user_id,
                'external_reference_id': self.external_reference_id,
                'custom_attributes': self.custom_attributes,
                'llm_start_time': datetime.now(timezone.utc),
                'language_model_id': kwargs.get('invocation_params').get('model_name')
            }
        except Exception as e:
            exception_message = (
                f"Error:\n"
                f"service name: athina-logger\n"
                f"file name: langchain_handler\n"
                f"method name: on_llm_start\n"
                f"{str(e)}"
            )
            print(exception_message)

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> None:
        try:
            run_info = self.runs.get(run_id, {})
            if not run_info:
                return
            llm_end_time = datetime.now(timezone.utc)
            run_info['response_time'] = round((
                (llm_end_time - run_info['llm_start_time']).total_seconds())*1000)

            for i in range(len(response.generations)):
                generation = response.generations[i][0]
                response_text = generation.text
                llm_output = response.llm_output
                token_usage = self._get_llm_usage(llm_output=llm_output, prompt=run_info['prompt'],
                                                response=response_text, language_model_id=run_info['language_model_id'], is_chat_model=run_info['is_chat_model'])

                run_info['response'] = response_text
                run_info['prompt_tokens'] = token_usage['prompt_tokens']
                run_info['completion_tokens'] = token_usage['completion_tokens']
                run_info['total_tokens'] = token_usage['total_tokens']

                # LOG TO API SERVER
                self._log_llm_response(run_info)
        except Exception as e:
            exception_message = (
                f"Error:\n"
                f"service name: athina-logger\n"
                f"file name: langchain_handler\n"
                f"method name: on_llm_end\n"
                f"{str(e)}"
            )
            print(exception_message)

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        **kwargs: Any,
    ) -> None:
        """Do nothing when tool starts."""
        pass

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        """Do nothing when agent takes a specific action."""
        pass

    def on_tool_end(
        self,
        output: str,
        observation_prefix: Optional[str] = None,
        llm_prefix: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Do nothing when tool ends."""
        pass

    def on_tool_error(
        self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> None:
        """Do nothing when tool outputs an error."""
        pass

    def on_text(self, text: str, **kwargs: Any) -> None:
        """Do nothing"""
        pass

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> None:
        """Do nothing"""
        pass

    def _get_llm_usage(self, llm_output: Dict, prompt, response, language_model_id, is_chat_model) -> Dict:
        """
        Fetch prompt tokens, completion tokens and total tokens from llm output
        """
        if llm_output is not None and 'token_usage' in llm_output:
            return {
                'prompt_tokens': llm_output['token_usage']['prompt_tokens'] if 'prompt_tokens' in llm_output['token_usage'] else None,
                'completion_tokens': llm_output['token_usage']['completion_tokens'] if 'completion_tokens' in llm_output['token_usage'] else None,
                'total_tokens': llm_output['token_usage']['total_tokens'] if 'total_tokens' in llm_output['token_usage'] else None,
            }
        else:
            if is_chat_model:
                prompt_tokens = self._get_prompt_tokens_chat_model(
                    prompt=prompt, language_model_id=language_model_id)
                completion_tokens = self._get_completion_tokens_chat_model(
                    response=response, language_model_id=language_model_id)
                if prompt_tokens is not None and completion_tokens is not None:
                    total_tokens = prompt_tokens + completion_tokens
                else:
                    total_tokens = None
            else:
                prompt_tokens = self._get_token_usage_completion_model(
                    text=prompt, language_model_id=language_model_id)
                completion_tokens = self._get_token_usage_completion_model(
                    text=response, language_model_id=language_model_id)
                if prompt_tokens is not None and completion_tokens is not None:
                    total_tokens = prompt_tokens + completion_tokens
                else:
                    total_tokens = None
        return {
            'prompt_tokens': prompt_tokens,
            'completion_tokens': completion_tokens,
            'total_tokens': total_tokens
        }

    def _get_prompt_tokens_chat_model(self, prompt: List[Dict[str, Any]], language_model_id: str):
        """
        gets the prompt tokens given the prompt
        """
        try:
            tokens = get_prompt_tokens_openai_chat_completion(
                prompt=prompt, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    def _get_completion_tokens_chat_model(self, response: str, language_model_id: str):
        """
        gets the completion tokens given the prompt response from the openai chat model completion
        """
        try:
            tokens = get_completion_tokens_openai_chat_completion(
                response=response, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    def _get_token_usage_completion_model(self, text: str, language_model_id: str):
        """
        gets the token usage given the prompt or prompt response for the openai completion model
        """
        try:
            tokens = get_token_usage_openai_completion(
                text=text, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    def _log_llm_response(self, run_info: Dict):
        """
        Logs the LLM response to athina
        """
        try:
            InferenceLogger.log_inference(prompt_slug=run_info['prompt_slug'], prompt=run_info['prompt'],
                                                        response=run_info['response'], language_model_id=run_info['language_model_id'],
                                                       prompt_tokens=run_info['prompt_tokens'], completion_tokens=run_info[
                                                           'completion_tokens'],
                                                       total_tokens=run_info['total_tokens'], response_time=run_info['response_time'],
                                                       environment=self.environment, context=run_info[
                'context'], user_query=run_info['user_query'],
                customer_id=run_info['customer_id'], session_id=run_info['session_id'],
                customer_user_id=run_info['customer_user_id'], external_reference_id=run_info['external_reference_id'],custom_attributes=run_info['custom_attributes'])

        except Exception as e:
            exception_message = (
                f"Error:\n"
                f"service name: athina-logger\n"
                f"file name: langchain_handler\n"
                f"method name: _log_llm_response\n"
                f"{str(e)}"
            )
            print(exception_message)

    def _convert_message_to_dict(self, message: BaseMessage) -> Dict[str, Any]:
        if isinstance(message, HumanMessage):
            message_dict = {'role': 'user', 'content': message.content}
        elif isinstance(message, AIMessage):
            message_dict = {'role': 'assistant', 'content': message.content}
        elif isinstance(message, SystemMessage):
            message_dict = {'role': 'system', 'content': message.content}
        elif isinstance(message, ChatMessage):
            message_dict = {'role': message.role, 'content': message.content}
        else:
            raise ValueError(f'Got unknown type {message}')
        if 'name' in message.additional_kwargs:
            message_dict['name'] = message.additional_kwargs['name']
        return message_dict

    def _create_message_dicts(
        self, messages: List[BaseMessage]
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        message_dicts = [self._convert_message_to_dict(m) for m in messages]
        return message_dicts



from typing import Dict, List, Any
import tiktoken
from ..constants import OPENAI_MODEL_ENCODINGS

# source: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb


def get_prompt_tokens_openai_chat_completion(prompt: List[Dict[str, Any]], language_model_id: str):
    """
    gets the prompt tokens given the prompt for the openai chat model completion
    """
    if prompt is None:
        raise ValueError('prompt is None')

    if language_model_id in {
        'gpt-3.5-turbo-0613',
        'gpt-3.5-turbo-16k-0613',
        'gpt-4-0613',
        'gpt-4-32k-0613',
    }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif 'gpt-3.5-turbo' in language_model_id:
        return get_prompt_tokens_openai_chat_completion(prompt=prompt, language_model_id='gpt-3.5-turbo-0613')
    elif 'gpt-4' in language_model_id:
        return get_prompt_tokens_openai_chat_completion(prompt=prompt, language_model_id='gpt-4-0613')
    else:
        raise ValueError(
            f'Language model {language_model_id} is not supported')

    try:
        encoding = tiktoken.get_encoding(OPENAI_MODEL_ENCODINGS[language_model_id])
    except KeyError:
        return None

    try:
        num_tokens = 0
        for message in prompt:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == 'name':
                    num_tokens += tokens_per_name
        num_tokens += 3
        return num_tokens
    except Exception as e:
        raise e


def get_completion_tokens_openai_chat_completion(response: str, language_model_id: str):
    """
    gets the completion tokens given the prompt response from the openai chat model completion
    """
    if response is None:
        raise ValueError('response is None')
    try:
        if language_model_id in {
            'gpt-3.5-turbo-0613',
            'gpt-3.5-turbo-16k-0613',
            'gpt-4-0613',
            'gpt-4-32k-0613',
        }:
            encoding = tiktoken.get_encoding(OPENAI_MODEL_ENCODINGS[language_model_id])
        elif 'gpt-3.5-turbo' in language_model_id:
            return get_completion_tokens_openai_chat_completion(response=response, language_model_id='gpt-3.5-turbo-0613')
        elif 'gpt-4' in language_model_id:
            return get_completion_tokens_openai_chat_completion(response=response, language_model_id='gpt-4-0613')
        else:
            raise ValueError(
                f'Language model {language_model_id} is not supported')
    except KeyError:
        return None

    try:
        tokens = len(encoding.encode(response))
        return tokens
    except Exception as e:
        raise e


def get_token_usage_openai_completion(text: str, language_model_id: str):
    """
    gets the token usage given the text and language_model_id for openai completion
    """
    if text is None:
        raise ValueError('text is None')
    try:
        encoding = tiktoken.get_encoding(OPENAI_MODEL_ENCODINGS[language_model_id])
    except KeyError:
        return None

    try:
        tokens = len(encoding.encode(text))
        return tokens
    except Exception as e:
        raise e




from typing import Optional
from ..api_key import AthinaApiKey
from ..constants import API_BASE_URL
from ..request_helper import RequestHelper


class UserFeedback(AthinaApiKey):
    """
    class for logging user feedback.
    """
    @staticmethod
    def log_user_feedback(
        external_reference_id: str,
        user_feedback: int,
        user_feedback_comment: Optional[str] = None,
    ) -> None:
        """
        logs the user feedback of the prompt run given the external reference id.
        """
        try:
            payload = {
                'external_reference_id': external_reference_id,
                'user_feedback': user_feedback,
                'user_feedback_comment': user_feedback_comment,
            }
            # Remove None fields from the payload
            payload = {k: v for k, v in payload.items() if v is not None}
            RequestHelper.make_patch_request(endpoint=f'{API_BASE_URL}/api/v1/prompt_run/user-feedback', payload=payload, headers={
                'athina-api-key': UserFeedback.get_api_key(),
            })
        except Exception as e:
            raise e



from abc import ABC, abstractmethod
from typing import Optional, Dict


class LogStreamInference(ABC):
    """
    abstract class for log stream inference.
    """

    def __init__(self,
                 prompt_slug: str,
                 response_time: Optional[int] = None,
                 context: Optional[Dict] = None,
                 environment: Optional[str] = 'production',
                 customer_id: Optional[str] = None,
                 customer_user_id: Optional[str] = None,
                 session_id: Optional[str] = None,
                 user_query: Optional[str] = None,
                 external_reference_id: Optional[str] = None,
                 custom_attributes: Optional[Dict] = None):
        """
        constructor for log stream inference
        :param prompt_slug: str - The slug of the prompt used for the inference.
        :param response_time: Optional[int] - The response time in milliseconds. Defaults to None.
        :param context: Optional[Dict] - A dictionary containing additional context information. Defaults to None.
        :param environment: Optional[str] - The environment in which the inference occurred. Defaults to production.
        :param customer_id: Optional[str] - The customer's unique identifier. Defaults to None.
        :param customer_user_id: Optional[str] - The customer's user identifier. Defaults to None.
        :param session_id: Optional[str] - The session identifier. Defaults to None.
        :param user_query: Optional[str] - The user's query or input. Defaults to None.
        :param external_reference_id: Optional[str] - The external reference id. Defaults to None.
        :param custom_attributes: Optional[Dict] - A dictionary containing custom attributes. Defaults to None.
        """
        self._prompt_slug = prompt_slug
        self._response_time = response_time
        self._context = context
        self._environment = environment
        self._customer_id = customer_id
        self._customer_user_id = customer_user_id
        self._session_id = session_id
        self._user_query = user_query
        self._external_reference_id = external_reference_id
        self._custom_attributes = custom_attributes

    @property
    def prompt_slug(self):
        return self._prompt_slug

    @prompt_slug.setter
    def prompt_slug(self, value):
        self._prompt_slug = value

    @property
    def response_time(self):
        return self._response_time

    @response_time.setter
    def response_time(self, value):
        self._response_time = value

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value):
        self._context = value

    @property
    def environment(self):
        return self._environment

    @environment.setter
    def environment(self, value):
        self._environment = value

    @property
    def customer_id(self):
        return self._customer_id

    @customer_id.setter
    def customer_id(self, value):
        self._customer_id = value

    @property
    def customer_user_id(self):
        return self._customer_user_id

    @customer_user_id.setter
    def customer_user_id(self, value):
        self._customer_user_id = value

    @property
    def session_id(self):
        return self._session_id

    @session_id.setter
    def session_id(self, value):
        self._session_id = value

    @property
    def user_query(self):
        return self._user_query

    @user_query.setter
    def user_query(self, value):
        self._user_query = value

    @property
    def external_reference_id(self):
        return self._external_reference_id

    @external_reference_id.setter
    def external_reference_id(self, value):
        self._external_reference_id = value

    @property
    def custom_attributes(self):
        return self._custom_attributes

    @custom_attributes.setter
    def custom_attributes(self, value):
        self._custom_attributes = value

    @abstractmethod
    def collect_stream_inference(self, response):
        """
        collects the inference from the log stream
        """
        pass

    @abstractmethod
    def collect_stream_inference_by_chunk(self, stream_chunk):
        """
        collects the inference from the log stream by chunk
        """
        pass

    @abstractmethod
    def log_stream_inference(self):
        """
        logs the stream inference to the athina api server
        """
        pass



from dataclasses import dataclass
from typing import Optional


@dataclass
class AthinaMeta:
    prompt_slug: Optional[str] = None
    context: Optional[dict] = None
    customer_id: Optional[dict] = None
    customer_user_id: Optional[dict] = None
    session_id: Optional[dict] = None
    user_query: Optional[dict] = None
    environment: Optional[dict] = None
    external_reference_id: Optional[dict] = None
    customer_id: Optional[str] = None
    customer_user_id: Optional[str] = None
    response_time: Optional[int] = None
    custom_attributes: Optional[dict] = None



# Athina Server Base Url
API_BASE_URL = 'https://log.athina.ai'
LOG_INFERENCE_URL = f'{API_BASE_URL}/api/v1/log/inference'

OPENAI_MODEL_ENCODINGS = {
    'gpt-3.5-turbo-0613': 'cl100k_base',
    'gpt-3.5-turbo-16k-0613': 'cl100k_base',
    'gpt-4-0613': 'cl100k_base',
    'gpt-4-32k-0613': 'cl100k_base',
    'text-davinci-003': 'p50k_base',
}



class CustomException(Exception):
    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message
        super().__init__(f"Error {status_code}: {message}")



import importlib
from dataclasses import dataclass
import datetime
import functools
import traceback
import threading
from typing import Callable, Optional, List
import time
from .athina_meta import AthinaMeta
from .inference_logger import InferenceLogger
from .api_key import AthinaApiKey
import openai


# Check OpenAI version
openai_version = openai.__version__
version_numbers = tuple(map(int, openai_version.split('.')))


def log_to_athina(result: dict, args: dict, athina_meta: AthinaMeta):
    try:
        prompt_slug = "default"
        context = None
        customer_id = None
        customer_user_id = None
        response_time_ms = None
        session_id = None
        user_query = None
        environment = "production"
        external_reference_id = None
        custom_attributes = None

        if athina_meta:
            prompt_slug = athina_meta.prompt_slug
            context = athina_meta.context
            response_time_ms = athina_meta.response_time
            customer_id = athina_meta.customer_id
            customer_user_id = athina_meta.customer_user_id
            session_id = athina_meta.session_id
            user_query = athina_meta.user_query
            environment = athina_meta.environment or "production"
            external_reference_id = athina_meta.external_reference_id
            custom_attributes = athina_meta.custom_attributes

        InferenceLogger.log_inference(
            prompt_slug=prompt_slug,
            prompt=args["messages"],
            language_model_id=args["model"],
            response=result,
            context=context,
            response_time=response_time_ms,
            customer_id=customer_id,
            customer_user_id=customer_user_id,
            session_id=session_id,
            user_query=user_query,
            environment=environment,
            external_reference_id=external_reference_id,
            custom_attributes=custom_attributes,
        )
    except Exception as e:
        print("Exception while logging to Athina: ", e)
        traceback.print_exc()


class OpenAiMiddleware:
    _athina_meta: Optional[AthinaMeta]
    _args: Optional[any]
    _kwargs: Optional[dict]

    def __init__(self):
        pass

    def _with_athina_logging(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Extract args from OpenAI call
            self._args = args
            self._kwargs = kwargs
            self._athina_meta = kwargs.pop("athina_meta", None)

            # Make the OpenAI call and measure response time
            start_time = time.time()
            openai_response = func(*self._args, **self._kwargs)
            end_time = time.time()
            response_time_ms = int((end_time - start_time) * 1000)

            # Return if no result was returned from OpenAI
            if openai_response is None:
                print("No result was returned from OpenAI")
                return openai_response

            # Construct the Athina Meta object and log to Athina
            try:
                # Construct the Athina Meta object
                if self._athina_meta is not None:
                    self._athina_meta.response_time = response_time_ms
                else:
                    self._athina_meta = AthinaMeta(
                        prompt_slug="default",
                        response_time=response_time_ms,
                        environment="default",
                    )

                # Log to Athina if it is not a streamed response
                if "stream" not in self._kwargs or not self._kwargs["stream"]:
                    api_thread = threading.Thread(
                        target=log_to_athina,
                        kwargs={
                            "result": openai_response if version_numbers < (1, 0, 0) else openai_response.model_dump(),
                            "args": self._kwargs,
                            "athina_meta": self._athina_meta,
                        },
                    )
                    api_thread.start()

                return openai_response

            except Exception as e:
                print("Exception in Athina logging: ", e)
                traceback.print_exc()
                return openai_response

        return wrapper

    # Apply the Athina logging wrapper to OpenAI methods
    def apply_athina(self, openai_instance=None):
        openai_version = openai.__version__
        version_numbers = tuple(map(int, openai_version.split('.')))

        if version_numbers < (1, 0, 0):
            ChatCompletion = importlib.import_module(
                "openai.api_resources").ChatCompletion
            openai_method_name = "create"
            openai_method = getattr(ChatCompletion, openai_method_name)

            # Override the create method with the Athina logging wrapper
            athina_method = self._with_athina_logging(openai_method)
            setattr(ChatCompletion, openai_method_name, athina_method)
        else:
            openai_method_name = "create"
            if openai_instance is not None:
                openai_method = getattr(
                    openai_instance.chat.completions, openai_method_name)

                # Override the chat.completions.create method with the Athina logging wrapper
                athina_method = self._with_athina_logging(openai_method)
                setattr(openai_instance.chat.completions,
                        openai_method_name, athina_method)


middleware = OpenAiMiddleware()
middleware.apply_athina()


if version_numbers > (1, 0, 0):
    # Monkey-patch the constructor of openai.OpenAI
    original_openai_constructor = importlib.import_module(
        "openai").OpenAI.__init__

    def new_openai_constructor(self, *args, **kwargs):
        original_openai_constructor(self, *args, **kwargs)
        middleware.apply_athina(self)

    importlib.import_module(
        "openai").OpenAI.__init__ = new_openai_constructor



import requests
from typing import List, Optional, Dict, Union, Any
from .constants import API_BASE_URL
from .api_key import AthinaApiKey
from .request_helper import RequestHelper


class InferenceLogger(AthinaApiKey):

    @staticmethod
    def log_inference(
        prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
        response: Optional[Any] = None,
        prompt_slug: Optional[str] = None,
        language_model_id: Optional[str] = None,
        environment: Optional[str] = 'production',
        functions: Optional[List[Dict]] = None,
        function_call_response: Optional[Any] = None,
        tools: Optional[Any] = None,
        tool_calls: Optional[Any] = None,
        external_reference_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        user_query: Optional[str] = None,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        response_time: Optional[int] = None,
        context: Optional[Dict] = None,
        expected_response: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        cost: Optional[float] = None,
    ) -> None:
        """
        logs the llm inference to athina
        """
        try:
            payload = {
                'prompt': prompt,
                'response': response,
                'prompt_slug': prompt_slug,
                'language_model_id': language_model_id,
                'functions': functions,
                'function_call_response': function_call_response,
                'tools': tools,
                'tool_calls': tool_calls,
                'response_time': response_time,
                'context': context,
                'environment': environment,
                'customer_id': str(customer_id) if customer_id is not None else None,
                'customer_user_id': str(customer_user_id) if customer_user_id is not None else None,
                'session_id': str(session_id) if session_id is not None else None,
                'user_query': str(user_query) if user_query is not None else None,
                'external_reference_id': str(external_reference_id) if external_reference_id is not None else None,
                'prompt_tokens': prompt_tokens,
                'completion_tokens': completion_tokens,
                'total_tokens': total_tokens,
                'expected_response': expected_response,
                'custom_attributes': custom_attributes,
                'cost': cost,
            }
            # Remove None fields from the payload
            payload = {k: v for k, v in payload.items() if v is not None}
            RequestHelper.make_post_request(endpoint=f'{API_BASE_URL}/api/v1/log/inference', payload=payload, headers={
                'athina-api-key': InferenceLogger.get_api_key(),
            })
        except Exception as e:
            raise e



import requests
from retrying import retry

from .exception.custom_exception import CustomException


class RequestHelper:
    """
    class to make requests to the athina api
    """
    @staticmethod
    @retry(wait_fixed=100, stop_max_attempt_number=2)
    def make_post_request(endpoint: str, payload: dict, headers: dict):
        try:
            response = requests.post(
                endpoint,
                json=payload,
                headers=headers,
            )
            if response.status_code != 200:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(
                    response.status_code, f'{error_message}: {details_message}')
        except requests.exceptions.RequestException as e:
            raise e
        except Exception as e:
            raise e

    @staticmethod
    @retry(wait_fixed=100, stop_max_attempt_number=2)
    def make_patch_request(endpoint: str, payload: dict, headers: dict):
        try:
            response = requests.patch(
                endpoint,
                json=payload,
                headers=headers,
            )
            if response.status_code != 200:
                response_json = response.json()
                error_message = response_json.get('error', 'Unknown Error')
                details_message = response_json.get(
                    'details', {}).get('message', 'No Details')
                raise CustomException(
                    response.status_code, f'{error_message}: {details_message}')
        except requests.exceptions.RequestException as e:
            raise e
        except Exception as e:
            raise e



from abc import ABC


class AthinaApiKey(ABC):
    _athina_api_key = None

    @classmethod
    def set_api_key(cls, api_key):
        cls._athina_api_key = api_key

    @classmethod
    def get_api_key(cls):
        return cls._athina_api_key
